#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import platform
import shutil
import sys

sys.path.append('bin')
from autojump_argparse import ArgumentParser  # noqa

SUPPORTED_SHELLS = ('bash', 'zsh', 'fish', 'tcsh')


def cp(src, dest, dryrun=False):
    print('copying file: %s -> %s' % (src, dest))
    if not dryrun:
        shutil.copy(src, dest)


def get_shell():
    return os.path.basename(os.getenv('SHELL', ''))


def mkdir(path, dryrun=False):
    print('creating directory:', path)
    if not dryrun and not os.path.exists(path):
        os.makedirs(path)


def modify_autojump_sh(etc_dir, share_dir, dryrun=False):
    """Append custom installation path to autojump.sh"""
    custom_install = '\
        \n# check custom install \
        \nif [ -s %s/autojump.${shell} ]; then \
        \n    source %s/autojump.${shell} \
        \nfi\n' % (share_dir, share_dir)

    with open(os.path.join(etc_dir, 'autojump.sh'), 'a') as f:
        f.write(custom_install)


def modify_autojump_lua(clink_dir, bin_dir, dryrun=False):
    """Prepend custom AUTOJUMP_BIN_DIR definition to autojump.lua"""
    custom_install = "local AUTOJUMP_BIN_DIR = \"%s\"\n" % bin_dir.replace(
        '\\',
        '\\\\')
    clink_file = os.path.join(clink_dir, 'autojump.lua')
    with open(clink_file, 'r') as f:
        original = f.read()
    with open(clink_file, 'w') as f:
        f.write(custom_install + original)


def parse_arguments():  # noqa
    if platform.system() == 'Windows':
        default_user_destdir = os.path.join(
            os.getenv('LOCALAPPDATA', ''),
            'autojump')
    else:
        default_user_destdir = os.path.join(
            os.path.expanduser('~'),
            '.autojump')
    default_user_prefix = ''
    default_user_zshshare = 'functions'
    default_system_destdir = '/'
    default_system_prefix = '/usr/local'
    default_system_zshshare = '/usr/share/zsh/site-functions'
    default_clink_dir = os.path.join(os.getenv('LOCALAPPDATA', ''), 'clink')

    parser = ArgumentParser(
        description='Installs autojump globally for root users, otherwise \
            installs in current user\'s home directory.')
    parser.add_argument(
        '-n', '--dryrun', action='store_true', default=False,
        help='simulate installation')
    parser.add_argument(
        '-f', '--force', action='store_true', default=False,
        help='skip root user, shell type, Python version checks')
    parser.add_argument(
        '-d', '--destdir', metavar='DIR', default=default_user_destdir,
        help='set destination to DIR')
    parser.add_argument(
        '-p', '--prefix', metavar='DIR', default=default_user_prefix,
        help='set prefix to DIR')
    parser.add_argument(
        '-z', '--zshshare', metavar='DIR', default=default_user_zshshare,
        help='set zsh share destination to DIR')
    parser.add_argument(
        '-c', '--clinkdir', metavar='DIR', default=default_clink_dir,
        help='set clink directory location to DIR (Windows only)')
    parser.add_argument(
        '-s', '--system', action='store_true', default=False,
        help='install system wide for all users')

    args = parser.parse_args()

    if not args.force:
        if sys.version_info[0] == 2 and sys.version_info[1] < 6:
            print('Python v2.6+ or v3.0+ required.', file=sys.stderr)
            sys.exit(1)
        if args.system:
            if platform.system() == 'Windows':
                print('System-wide installation is not supported on Windows.',
                      file=sys.stderr)
                sys.exit(1)
            elif os.geteuid() != 0:
                print('Please rerun as root for system-wide installation.',
                      file=sys.stderr)
                sys.exit(1)

        if platform.system() != 'Windows' \
                and get_shell() not in SUPPORTED_SHELLS:
            print('Unsupported shell: %s' % os.getenv('SHELL'),
                  file=sys.stderr)
            sys.exit(1)

    if args.destdir != default_user_destdir \
            or args.prefix != default_user_prefix \
            or args.zshshare != default_user_zshshare:
        args.custom_install = True
    else:
        args.custom_install = False

    if args.system:
        if args.custom_install:
            print('Custom paths incompatible with --system option.',
                  file=sys.stderr)
            sys.exit(1)

        args.destdir = default_system_destdir
        args.prefix = default_system_prefix
        args.zshshare = default_system_zshshare

    return args


def show_post_installation_message(etc_dir, share_dir, bin_dir):
    if platform.system() == 'Windows':
        print('\nPlease manually add %s to your user path' % bin_dir)
    else:
        if get_shell() == 'fish':
            aj_shell = '%s/autojump.fish' % share_dir
            source_msg = 'if test -f %s; . %s; end' % (aj_shell, aj_shell)
            rcfile = '~/.config/fish/config.fish'
        else:
            aj_shell = '%s/autojump.sh' % etc_dir
            source_msg = '[[ -s %s ]] && source %s' % (aj_shell, aj_shell)

            if platform.system() == 'Darwin' and get_shell() == 'bash':
                rcfile = '~/.profile'
            else:
                rcfile = '~/.%src' % get_shell()

        print('\nPlease manually add the following line(s) to %s:' % rcfile)
        print('\n\t' + source_msg)
        if get_shell() == 'zsh':
            print('\n\tautoload -U compinit && compinit -u')

    print('\nPlease restart terminal(s) before running autojump.\n')


def main(args):
    if args.dryrun:
        print('Installing autojump to %s (DRYRUN)...' % args.destdir)
    else:
        print('Installing autojump to %s ...' % args.destdir)

    bin_dir = os.path.join(args.destdir, args.prefix, 'bin')
    etc_dir = os.path.join(args.destdir, 'etc', 'profile.d')
    doc_dir = os.path.join(args.destdir, args.prefix, 'share', 'man', 'man1')
    share_dir = os.path.join(args.destdir, args.prefix, 'share', 'autojump')
    zshshare_dir = os.path.join(args.destdir, args.zshshare)

    mkdir(bin_dir, args.dryrun)
    mkdir(doc_dir, args.dryrun)
    mkdir(etc_dir, args.dryrun)
    mkdir(share_dir, args.dryrun)

    cp('./bin/autojump', bin_dir, args.dryrun)
    cp('./bin/autojump_argparse.py', bin_dir, args.dryrun)
    cp('./bin/autojump_data.py', bin_dir, args.dryrun)
    cp('./bin/autojump_match.py', bin_dir, args.dryrun)
    cp('./bin/autojump_utils.py', bin_dir, args.dryrun)
    cp('./bin/icon.png', share_dir, args.dryrun)
    cp('./docs/autojump.1', doc_dir, args.dryrun)

    if platform.system() == 'Windows':
        cp('./bin/autojump.lua', args.clinkdir, args.dryrun)
        cp('./bin/autojump.bat', bin_dir, args.dryrun)
        cp('./bin/j.bat', bin_dir, args.dryrun)
        cp('./bin/jc.bat', bin_dir, args.dryrun)
        cp('./bin/jo.bat', bin_dir, args.dryrun)
        cp('./bin/jco.bat', bin_dir, args.dryrun)

        if args.custom_install:
            modify_autojump_lua(args.clinkdir, bin_dir, args.dryrun)
    else:
        mkdir(etc_dir, args.dryrun)
        mkdir(share_dir, args.dryrun)
        mkdir(zshshare_dir, args.dryrun)

        cp('./bin/autojump.sh', etc_dir, args.dryrun)
        cp('./bin/autojump.bash', share_dir, args.dryrun)
        cp('./bin/autojump.fish', share_dir, args.dryrun)
        cp('./bin/autojump.zsh', share_dir, args.dryrun)
        cp('./bin/_j', zshshare_dir, args.dryrun)

        if args.custom_install:
            modify_autojump_sh(etc_dir, share_dir, args.dryrun)

    show_post_installation_message(etc_dir, share_dir, bin_dir)


if __name__ == '__main__':
    sys.exit(main(parse_arguments()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import errno
import os
import platform
import re
import shutil
import sys
import unicodedata
from itertools import islice

if sys.version_info[0] == 3:
    imap = map
    os.getcwdu = os.getcwd
else:
    from itertools import imap


def create_dir(path):
    """Creates a directory atomically."""
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def encode_local(string):
    """Converts string into user's preferred encoding."""
    if is_python3():
        return string
    return string.encode(sys.getfilesystemencoding() or 'utf-8')


def first(xs):
    it = iter(xs)
    try:
        if is_python3():
            return it.__next__()
        return it.next()
    except StopIteration:
        return None


def get_tab_entry_info(entry, separator):
    """
    Given a tab entry in the following format return needle, index, and path:

        [needle]__[index]__[path]
    """
    needle, index, path = None, None, None

    match_needle = re.search(r'(.*?)' + separator, entry)
    match_index = re.search(separator + r'([0-9]{1})', entry)
    match_path = re.search(
        separator + r'[0-9]{1}' + separator + r'(.*)',
        entry)

    if match_needle:
        needle = match_needle.group(1)

    if match_index:
        index = int(match_index.group(1))

    if match_path:
        path = match_path.group(1)

    return needle, index, path


def get_pwd():
    try:
        return os.getcwdu()
    except OSError:
        print('Current directory no longer exists.', file=sys.stderr)
        raise


def has_uppercase(string):
    if is_python3():
        return any(ch.isupper() for ch in string)
    return any(unicodedata.category(c) == 'Lu' for c in unicode(string))


def in_bash():
    return 'bash' in os.getenv('SHELL')


def is_autojump_sourced():
    return '1' == os.getenv('AUTOJUMP_SOURCED')


def is_python2():
    return sys.version_info[0] == 2


def is_python3():
    return sys.version_info[0] == 3


def is_linux():
    return platform.system() == 'Linux'


def is_osx():
    return platform.system() == 'Darwin'


def is_windows():
    return platform.system() == 'Windows'


def last(xs):
    it = iter(xs)
    tmp = None
    try:
        if is_python3():
            while True:
                tmp = it.__next__()
        else:
            while True:
                tmp = it.next()
    except StopIteration:
        return tmp


def move_file(src, dst):
    """
    Atomically move file.

    Windows does not allow for atomic file renaming (which is used by
    os.rename / shutil.move) so destination paths must first be deleted.
    """
    if is_windows() and os.path.exists(dst):
        # raises exception if file is in use on Windows
        os.remove(dst)
    shutil.move(src, dst)


def print_entry(entry):
    print_local('%.1f:\t%s' % (entry.weight, entry.path))


def print_local(string):
    print(encode_local(string))


def print_tab_menu(needle, tab_entries, separator):
    """
    Prints the tab completion menu according to the following format:

        [needle]__[index]__[possible_match]

    The needle (search pattern) and index are necessary to recreate the results
    on subsequent calls.
    """
    for i, entry in enumerate(tab_entries):
        print_local(
            '%s%s%d%s%s' % (
                needle,
                separator,
                i + 1,
                separator,
                entry.path))


def sanitize(directories):
    # edge case to allow '/' as a valid path
    clean = lambda x: unico(x) if x == os.sep else unico(x).rstrip(os.sep)
    return list(imap(clean, directories))


def second(xs):
    it = iter(xs)
    try:
        if is_python2():
            it.next()
            return it.next()
        elif is_python3():
            next(it)
            return next(it)
    except StopIteration:
        return None


def surround_quotes(string):
    """
    Bash has problems dealing with certain paths so we're surrounding all
    path outputs with quotes.
    """
    if in_bash() and string:
        # Python 2.6 requres field numbers
        return '"{0}"'.format(string)
    return string


def take(n, iterable):
    """Return first n items of an iterable."""
    return islice(iterable, n)


def unico(string):
    """Converts into Unicode string."""
    if is_python2() and not isinstance(string, unicode):
        return unicode(string, encoding='utf-8', errors='replace')
    return string



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys

import mock
import pytest

sys.path.append(os.path.join(os.getcwd(), 'bin'))  # noqa
import autojump_utils
from autojump_utils import encode_local
from autojump_utils import first
from autojump_utils import get_tab_entry_info
from autojump_utils import has_uppercase
from autojump_utils import in_bash
from autojump_utils import is_python3
from autojump_utils import last
from autojump_utils import sanitize
from autojump_utils import second
from autojump_utils import surround_quotes
from autojump_utils import take
from autojump_utils import unico


if is_python3():
    os.getcwdu = os.getcwd
    xrange = range


def u(string):
    """
    This is a unicode() wrapper since u'string' is a Python3 compiler error.
    """
    if is_python3():
        return string
    return unicode(string, encoding='utf-8', errors='strict')


# strings
@pytest.mark.skipif(is_python3(), reason='Unicode sucks.')
@mock.patch.object(sys, 'getfilesystemencoding', return_value='ascii')
def test_encode_local_ascii(_):
    assert encode_local(u('foo')) == b'foo'


@pytest.mark.skipif(is_python3(), reason='Unicode sucks.')
@pytest.mark.xfail(reason='disabled due to pytest bug: https://bitbucket.org/hpk42/pytest/issue/534/pytest-fails-to-catch-unicodedecodeerrors')  # noqa
@mock.patch.object(sys, 'getfilesystemencoding', return_value='ascii')
def test_encode_local_ascii_fails(_):
    with pytest.raises(UnicodeDecodeError):
        encode_local(u('日本語'))


@pytest.mark.skipif(is_python3(), reason='Unicode sucks.')
@mock.patch.object(sys, 'getfilesystemencoding', return_value=None)
def test_encode_local_empty(_):
    assert encode_local(b'foo') == u('foo')


@pytest.mark.skipif(is_python3(), reason='Unicode sucks.')
@mock.patch.object(sys, 'getfilesystemencoding', return_value='utf-8')
def test_encode_local_unicode(_):
    assert encode_local(b'foo') == u('foo')
    assert encode_local(u('foo')) == u('foo')


def test_has_uppercase():
    assert has_uppercase('Foo')
    assert has_uppercase('foO')
    assert not has_uppercase('foo')
    assert not has_uppercase('')


@mock.patch.object(autojump_utils, 'in_bash', return_value=True)
def test_surround_quotes_in_bash(_):
    assert surround_quotes('foo') == '"foo"'


@mock.patch.object(autojump_utils, 'in_bash', return_value=False)
def test_dont_surround_quotes_not_in_bash(_):
    assert surround_quotes('foo') == 'foo'


def test_sanitize():
    assert sanitize([]) == []
    assert sanitize([r'/foo/bar/', r'/']) == [u('/foo/bar'), u('/')]


@pytest.mark.skipif(is_python3(), reason='Unicode sucks.')
def test_unico():
    assert unico(str('blah')) == u('blah')
    assert unico(str('日本語')) == u('日本語')
    assert unico(u('でもおれは中国人だ。')) == u('でもおれは中国人だ。')


# iteration
def test_first():
    assert first(xrange(5)) == 0
    assert first([]) is None


def test_second():
    assert second(xrange(5)) == 1
    assert second([]) is None


def test_last():
    assert last(xrange(4)) == 3
    assert last([]) is None


def test_take():
    assert list(take(1, xrange(3))) == [0]
    assert list(take(2, xrange(3))) == [0, 1]
    assert list(take(4, xrange(3))) == [0, 1, 2]
    assert list(take(10, [])) == []


# environment variables
def test_in_bash():
    for path in ['/bin/bash', '/usr/bin/bash']:
        os.environ['SHELL'] = path
        assert in_bash()

    for path in ['/bin/zsh', '/usr/bin/zsh']:
        os.environ['SHELL'] = '/usr/bin/zsh'
        assert not in_bash()


# helper functions
def test_get_needle():
    assert get_tab_entry_info('foo__', '__') == ('foo', None, None)


def test_get_index():
    assert get_tab_entry_info('foo__2', '__') == ('foo', 2, None)


def test_get_path():
    assert get_tab_entry_info('foo__3__/foo/bar', '__') \
        == ('foo', 3, '/foo/bar')


def test_get_none():
    assert get_tab_entry_info('gibberish content', '__') == (None, None, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import shutil
import sys
from codecs import open
from collections import namedtuple
from tempfile import NamedTemporaryFile
from time import time

from autojump_utils import create_dir
from autojump_utils import is_osx
from autojump_utils import is_python3
from autojump_utils import move_file
from autojump_utils import unico


if sys.version_info[0] == 3:
    ifilter = filter
    imap = map
else:
    from itertools import ifilter  # noqa
    from itertools import imap  # noqa


BACKUP_THRESHOLD = 24 * 60 * 60
Entry = namedtuple('Entry', ['path', 'weight'])


def dictify(entries):
    """
    Converts a list of entries into a dictionary where
        key = path
        value = weight
    """
    result = {}
    for entry in entries:
        result[entry.path] = entry.weight
    return result


def entriefy(data):
    """Converts a dictionary into an iterator of entries."""
    convert = lambda tup: Entry(*tup)
    if is_python3():
        return map(convert, data.items())
    return imap(convert, data.iteritems())


def load(config):
    """Returns a dictonary (key=path, value=weight) loaded from data file."""
    xdg_aj_home = os.path.join(
        os.path.expanduser('~'),
        '.local',
        'share',
        'autojump')

    if is_osx() and os.path.exists(xdg_aj_home):
        migrate_osx_xdg_data(config)

    if not os.path.exists(config['data_path']):
        return {}

    # example: u'10.0\t/home/user\n' -> ['10.0', u'/home/user']
    parse = lambda line: line.strip().split('\t')

    correct_length = lambda x: len(x) == 2

    # example: ['10.0', u'/home/user'] -> (u'/home/user', 10.0)
    tupleize = lambda x: (x[1], float(x[0]))

    try:
        with open(
                config['data_path'],
                'r', encoding='utf-8',
                errors='replace') as f:
            return dict(
                imap(
                    tupleize,
                    ifilter(correct_length, imap(parse, f))))
    except (IOError, EOFError):
        return load_backup(config)


def load_backup(config):
    if os.path.exists(config['backup_path']):
        move_file(config['backup_path'], config['data_path'])
        return load(config)
    return {}


def migrate_osx_xdg_data(config):
    """
    Older versions incorrectly used Linux XDG_DATA_HOME paths on OS X. This
    migrates autojump files from ~/.local/share/autojump to ~/Library/autojump
    """
    assert is_osx(), 'This function should only be run on OS X.'

    xdg_data_home = os.path.join(os.path.expanduser('~'), '.local', 'share')
    xdg_aj_home = os.path.join(xdg_data_home, 'autojump')
    data_path = os.path.join(xdg_aj_home, 'autojump.txt')
    backup_path = os.path.join(xdg_aj_home, 'autojump.txt.bak')

    if os.path.exists(data_path):
        move_file(data_path, config['data_path'])
    if os.path.exists(backup_path):
        move_file(backup_path, config['backup_path'])

    # cleanup
    shutil.rmtree(xdg_aj_home)
    if len(os.listdir(xdg_data_home)) == 0:
        shutil.rmtree(xdg_data_home)


def save(config, data):
    """Save data and create backup, creating a new data file if necessary."""
    create_dir(os.path.dirname(config['data_path']))

    # atomically save by writing to temporary file and moving to destination
    try:
        temp = NamedTemporaryFile(delete=False)
        # Windows cannot reuse the same open file name
        temp.close()

        with open(temp.name, 'w', encoding='utf-8', errors='replace') as f:
            for path, weight in data.items():
                f.write(unico('%s\t%s\n' % (weight, path)))

            f.flush()
            os.fsync(f)
    except IOError as ex:
        print('Error saving autojump data (disk full?)' % ex, file=sys.stderr)
        sys.exit(1)

    # move temp_file -> autojump.txt
    move_file(temp.name, config['data_path'])

    # create backup file if it doesn't exist or is older than BACKUP_THRESHOLD
    if not os.path.exists(config['backup_path']) or \
            (time() - os.path.getmtime(config['backup_path']) > BACKUP_THRESHOLD):  # noqa
        shutil.copy(config['data_path'], config['backup_path'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys

import pytest

sys.path.append(os.path.join(os.getcwd(), 'bin'))  # noqa
from autojump_data import Entry
from autojump_match import match_anywhere
from autojump_match import match_consecutive


class TestMatchAnywhere(object):

    entry1 = Entry('/foo/bar/baz', 10)
    entry2 = Entry('/baz/foo/bar', 10)
    entry3 = Entry('/foo/baz', 10)
    entry4 = Entry('/中/zhong/国/guo', 10)
    entry5 = Entry('/is\'t/this/a/b*tchin/edge/case?', 10)
    win_entry1 = Entry('C:\\foo\\bar\\baz', 10)
    win_entry2 = Entry('D:\Program Files (x86)\GIMP', 10)
    win_entry3 = Entry('C:\Windows\System32', 10)

    @pytest.fixture
    def haystack(self):
        return [
            self.entry1,
            self.entry2,
            self.entry3,
            self.entry4,
            self.entry5,
        ]

    @pytest.fixture
    def windows_haystack(self):
        return [self.win_entry1, self.win_entry2, self.win_entry3]

    def test_single_needle(self, haystack):
        assert list(match_anywhere(['bar'], haystack)) == [self.entry1, self.entry2]

    def test_consecutive(self, haystack):
        assert list(match_anywhere(['foo', 'bar'], haystack)) \
            == [self.entry1, self.entry2]
        assert list(match_anywhere(['bar', 'foo'], haystack)) == []

    def test_skip(self, haystack):
        assert list(match_anywhere(['baz', 'bar'], haystack)) == [self.entry2]
        assert list(match_anywhere(['中', '国'], haystack)) == [self.entry4]

    def test_ignore_case(self, haystack):
        assert list(match_anywhere(['bAz', 'bAR'], haystack, ignore_case=True)) \
            == [self.entry2]

    def test_backslashes_for_windows_paths(self, windows_haystack):
        # https://github.com/wting/autojump/issues/281
        assert list(match_anywhere(['foo', 'baz'], windows_haystack)) \
            == [self.win_entry1]
        assert list(match_anywhere(['program', 'gimp'], windows_haystack, True)) \
            == [self.win_entry2]
        assert list(match_anywhere(['win', '32'], windows_haystack, True)) \
            == [self.win_entry3]

    def test_wildcard_in_needle(self, haystack):
        # https://github.com/wting/autojump/issues/402
        assert list(match_anywhere(['*', 'this'], haystack)) == []
        assert list(match_anywhere(['this', '*'], haystack)) == [self.entry5]


class TestMatchConsecutive(object):

    entry1 = Entry('/foo/bar/baz', 10)
    entry2 = Entry('/baz/foo/bar', 10)
    entry3 = Entry('/foo/baz', 10)
    entry4 = Entry('/中/zhong/国/guo', 10)
    entry5 = Entry('/日/本', 10)
    entry6 = Entry('/is\'t/this/a/b*tchin/edge/case?', 10)
    win_entry1 = Entry('C:\Foo\Bar\Baz', 10)
    win_entry2 = Entry('D:\Program Files (x86)\GIMP', 10)
    win_entry3 = Entry('C:\Windows\System32', 10)

    @pytest.fixture
    def haystack(self):
        return [
            self.entry1,
            self.entry2,
            self.entry3,
            self.entry4,
            self.entry5,
        ]

    @pytest.fixture
    def windows_haystack(self):
        return [self.win_entry1, self.win_entry2, self.win_entry3]

    def test_single_needle(self, haystack):
        assert list(match_consecutive(['baz'], haystack)) == [self.entry1, self.entry3]
        assert list(match_consecutive(['本'], haystack)) == [self.entry5]

    def test_consecutive(self, haystack):
        assert list(match_consecutive(['bar', 'baz'], haystack)) == [self.entry1]
        assert list(match_consecutive(['foo', 'bar'], haystack)) == [self.entry2]
        assert list(match_consecutive(['国', 'guo'], haystack)) == [self.entry4]
        assert list(match_consecutive(['bar', 'foo'], haystack)) == []

    def test_ignore_case(self, haystack):
        assert list(match_consecutive(['FoO', 'bAR'], haystack, ignore_case=True)) \
            == [self.entry2]

    def test_windows_ignore_case(self, windows_haystack):
        assert list(match_consecutive(['gimp'], windows_haystack, True)) == [self.win_entry2]

    @pytest.mark.xfail(reason='https://github.com/wting/autojump/issues/418')
    def test_backslashes_for_windows_paths(self, windows_haystack):
        assert list(match_consecutive(['program', 'gimp'], windows_haystack, True)) \
            == [self.win_entry2]

    @pytest.mark.xfail(reason='https://github.com/wting/autojump/issues/418')
    def test_foo_bar_baz(self, windows_haystack):
        assert list(match_consecutive(['bar', 'baz'], windows_haystack, ignore_case=True)) \
            == [self.win_entry1]

    @pytest.mark.xfail(reason='https://github.com/wting/autojump/issues/402')
    def test_thing(self, windows_haystack):
        assert list(match_consecutive(['win', '32'], windows_haystack, True)) \
            == [self.win_entry3]

    @pytest.mark.xfail(reason='https://github.com/wting/autojump/issues/402')
    def test_wildcard_in_needle(self, haystack):
        assert list(match_consecutive(['*', 'this'], haystack)) == []
        assert list(match_consecutive(['*', 'edge', 'case'], haystack)) == [self.entry6]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import platform
import shutil
import sys

sys.path.append('bin')
from autojump_argparse import ArgumentParser  # noqa


def is_empty_dir(path):
    """
    Checks if any files are present within a directory and all sub-directories.
    """
    for _, _, files in os.walk(path):
        if files:
            return False
    return True


def parse_arguments():
    default_clink_dir = os.path.join(os.getenv('LOCALAPPDATA', ''), 'clink')

    parser = ArgumentParser(
        description='Uninstalls autojump.')
    parser.add_argument(
        '-n', '--dryrun', action='store_true', default=False,
        help='simulate installation')
    parser.add_argument(
        '-u', '--userdata', action='store_true', default=False,
        help='delete user data')
    parser.add_argument(
        '-d', '--destdir', metavar='DIR',
        help='custom destdir')
    parser.add_argument(
        '-p', '--prefix', metavar='DIR', default='',
        help='custom prefix')
    parser.add_argument(
        '-z', '--zshshare', metavar='DIR', default='functions',
        help='custom zshshare')
    parser.add_argument(
        '-c', '--clinkdir', metavar='DIR', default=default_clink_dir)

    return parser.parse_args()


def remove_custom_installation(args, dryrun=False):
    if not args.destdir:
        return

    bin_dir = os.path.join(args.destdir, args.prefix, 'bin')
    doc_dir = os.path.join(args.destdir, args.prefix, 'share', 'man', 'man1')
    etc_dir = os.path.join(args.destdir, 'etc', 'profile.d')
    share_dir = os.path.join(args.destdir, args.prefix, 'share', 'autojump')
    zshshare_dir = os.path.join(args.destdir, args.zshshare)

    if not os.path.exists(share_dir):
        return

    print('\nFound custom installation...')
    rm(os.path.join(bin_dir, 'autojump'), dryrun)
    rm(os.path.join(bin_dir, 'autojump_data.py'), dryrun)
    rm(os.path.join(bin_dir, 'autojump_utils.py'), dryrun)
    rm(os.path.join(bin_dir, 'autojump_argparse.py'), dryrun)
    if platform.system() == 'Windows':
        if os.path.exists(args.clinkdir):
            rm(os.path.join(args.clinkdir, 'autojump.lua'), dryrun)
        rm(os.path.join(bin_dir, 'autojump.bat'), dryrun)
        rm(os.path.join(bin_dir, 'j.bat'), dryrun)
        rm(os.path.join(bin_dir, 'jc.bat'), dryrun)
        rm(os.path.join(bin_dir, 'jco.bat'), dryrun)
        rm(os.path.join(bin_dir, 'jo.bat'), dryrun)
    else:
        rm(os.path.join(etc_dir, 'autojump.sh'), dryrun)
        rm(os.path.join(share_dir, 'autojump.bash'), dryrun)
        rm(os.path.join(share_dir, 'autojump.fish'), dryrun)
        rm(os.path.join(share_dir, 'autojump.tcsh'), dryrun)
        rm(os.path.join(share_dir, 'autojump.zsh'), dryrun)
        rm(os.path.join(zshshare_dir, '_j'), dryrun)
    rmdir(share_dir, dryrun)
    rm(os.path.join(doc_dir, 'autojump.1'), dryrun)

    if is_empty_dir(args.destdir):
        rmdir(args.destdir, dryrun)


def remove_system_installation(dryrun=False):
    default_destdir = '/'
    default_prefix = '/usr/local'
    default_zshshare = '/usr/share/zsh/site-functions'

    bin_dir = os.path.join(default_destdir, default_prefix, 'bin')
    doc_dir = os.path.join(
        default_destdir,
        default_prefix,
        'share',
        'man',
        'man1')
    etc_dir = os.path.join(default_destdir, 'etc', 'profile.d')
    share_dir = os.path.join(
        default_destdir,
        default_prefix,
        'share',
        'autojump')
    zshshare_dir = os.path.join(default_destdir, default_zshshare)

    if not os.path.exists(share_dir):
        return

    print('\nFound system installation...')

    if os.geteuid() != 0:
        print('Please rerun as root for system-wide uninstall, skipping...',
              file=sys.stderr)
        return

    rm(os.path.join(bin_dir, 'autojump'), dryrun)
    rm(os.path.join(bin_dir, 'autojump_data.py'), dryrun)
    rm(os.path.join(bin_dir, 'autojump_utils.py'), dryrun)
    rm(os.path.join(etc_dir, 'autojump.sh'), dryrun)
    rm(os.path.join(share_dir, 'autojump.bash'), dryrun)
    rm(os.path.join(share_dir, 'autojump.fish'), dryrun)
    rm(os.path.join(share_dir, 'autojump.tcsh'), dryrun)
    rm(os.path.join(share_dir, 'autojump.zsh'), dryrun)
    rm(os.path.join(zshshare_dir, '_j'), dryrun)
    rmdir(share_dir, dryrun)
    rm(os.path.join(doc_dir, 'autojump.1'), dryrun)


def remove_user_data(dryrun=False):
    if platform.system() == 'Darwin':
        data_home = os.path.join(
            os.path.expanduser('~'),
            'Library',
            'autojump')
    elif platform.system() == 'Windows':
        data_home = os.path.join(
            os.getenv('APPDATA'),
            'autojump')
    else:
        data_home = os.getenv(
            'XDG_DATA_HOME',
            os.path.join(
                os.path.expanduser('~'),
                '.local',
                'share',
                'autojump'))

    if os.path.exists(data_home):
        print('\nFound user data...')
        rmdir(data_home, dryrun)


def remove_user_installation(dryrun=False):
    if platform.system() == 'Windows':
        default_destdir = os.path.join(os.getenv('LOCALAPPDATA', ''),
                                       'autojump')
        clink_dir = os.path.join(os.getenv('LOCALAPPDATA', ''), 'clink')
    else:
        default_destdir = os.path.join(os.path.expanduser('~'), '.autojump')

    if os.path.exists(default_destdir):
        print('\nFound user installation...')
        rmdir(default_destdir, dryrun)
        if platform.system() == 'Windows' and os.path.exists(clink_dir):
            rm(os.path.join(clink_dir, 'autojump.lua'), dryrun)


def rm(path, dryrun):
    if os.path.exists(path):
        print('deleting file:', path)
        if not dryrun:
            os.remove(path)


def rmdir(path, dryrun):
    if os.path.exists(path):
        print('deleting directory:', path)
        if not dryrun:
            shutil.rmtree(path)


def main(args):
    if args.dryrun:
        print('Uninstalling autojump (DRYRUN)...')
    else:
        print('Uninstalling autojump...')

    remove_user_installation(args.dryrun)
    remove_system_installation(args.dryrun)
    remove_custom_installation(args, args.dryrun)
    if args.userdata:
        remove_user_data(args.dryrun)


if __name__ == '__main__':
    sys.exit(main(parse_arguments()))



# -*- coding: utf-8 -*-
"""
IPython autojump magic

Written by keith hughitt <keith.hughitt@gmail.com>, based on an earlier
version by Mario Pastorelli <pastorelli.mario@gmail.com>.

To install, create a new IPython user profile by running:

    ipython profile create

And copy this file into the "startup" folder of your new profile (e.g.
"$HOME/.config/ipython/profile_default/startup/").

@TODO: extend %cd to call "autojump -a"
"""
from subprocess import PIPE
from subprocess import Popen

from IPython.core.magic import register_line_magic

ip = get_ipython()  # noqa


@register_line_magic
def j(path):
    cmd = ['autojump'] + path.split()
    newpath = Popen(
        cmd,
        stdout=PIPE,
        shell=False).communicate()[0].strip()

    if newpath:
        ip.magic('cd %s' % newpath.decode('utf-8'))

# remove from namespace
del j



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
from difflib import SequenceMatcher

from autojump_utils import is_python3
from autojump_utils import last


if is_python3():  # pragma: no cover
    ifilter = filter
    imap = map
    os.getcwdu = os.getcwd
else:
    from itertools import ifilter
    from itertools import imap


def match_anywhere(needles, haystack, ignore_case=False):
    """
    Matches needles anywhere in the path as long as they're in the same (but
    not necessary consecutive) order.

    For example:
        needles = ['foo', 'baz']
        regex needle = r'.*foo.*baz.*'
        haystack = [
            (path='/foo/bar/baz', weight=10),
            (path='/baz/foo/bar', weight=10),
            (path='/foo/baz', weight=10),
        ]

        result = [
            (path='/moo/foo/baz', weight=10),
            (path='/foo/baz', weight=10),
        ]
    """
    regex_needle = '.*' + '.*'.join(imap(re.escape, needles)) + '.*'
    regex_flags = re.IGNORECASE | re.UNICODE if ignore_case else re.UNICODE
    found = lambda haystack: re.search(
        regex_needle,
        haystack.path,
        flags=regex_flags,
    )
    return ifilter(found, haystack)


def match_consecutive(needles, haystack, ignore_case=False):
    """
    Matches consecutive needles at the end of a path.

    For example:
        needles = ['foo', 'baz']
        haystack = [
            (path='/foo/bar/baz', weight=10),
            (path='/foo/baz/moo', weight=10),
            (path='/moo/foo/baz', weight=10),
            (path='/foo/baz', weight=10),
        ]

        # We can't actually use re.compile because of re.UNICODE
        regex_needle = re.compile(r'''
            foo     # needle #1
            [^/]*   # all characters except os.sep zero or more times
            /       # os.sep
            [^/]*   # all characters except os.sep zero or more times
            baz     # needle #2
            [^/]*   # all characters except os.sep zero or more times
            $       # end of string
            ''')

        result = [
            (path='/moo/foo/baz', weight=10),
            (path='/foo/baz', weight=10),
        ]
    """
    regex_no_sep = '[^' + os.sep + ']*'
    regex_no_sep_end = regex_no_sep + '$'
    regex_one_sep = regex_no_sep + os.sep + regex_no_sep
    regex_needle = regex_one_sep.join(imap(re.escape, needles)) + regex_no_sep_end
    regex_flags = re.IGNORECASE | re.UNICODE if ignore_case else re.UNICODE
    found = lambda entry: re.search(
        regex_needle,
        entry.path,
        flags=regex_flags,
    )
    return ifilter(found, haystack)


def match_fuzzy(needles, haystack, ignore_case=False, threshold=0.6):
    """
    Performs an approximate match with the last needle against the end of
    every path past an acceptable threshold.

    For example:
        needles = ['foo', 'bar']
        haystack = [
            (path='/foo/bar/baz', weight=11),
            (path='/foo/baz/moo', weight=10),
            (path='/moo/foo/baz', weight=10),
            (path='/foo/baz', weight=10),
            (path='/foo/bar', weight=10),
        ]

    result = [
            (path='/foo/bar/baz', weight=11),
            (path='/moo/foo/baz', weight=10),
            (path='/foo/baz', weight=10),
            (path='/foo/bar', weight=10),
        ]

    This is a weak heuristic and used as a last resort to find matches.
    """
    end_dir = lambda path: last(os.path.split(path))
    if ignore_case:
        needle = last(needles).lower()
        match_percent = lambda entry: SequenceMatcher(
            a=needle,
            b=end_dir(entry.path.lower())).ratio()
    else:
        needle = last(needles)
        match_percent = lambda entry: SequenceMatcher(
            a=needle,
            b=end_dir(entry.path)).ratio()
    meets_threshold = lambda entry: match_percent(entry) >= threshold
    return ifilter(meets_threshold, haystack)
